// MiniTrk by CyanPhase
//
// A pretty rudimentary little tracker that shows how to do some
// stuff. Use at your own risk, and stuff.
// email stuff to blakee@rovoscape.com
//
// It has really nothing, not even volume ramping (sample anti-click).
// however it does use CSI for interpolation. No tracker effects are
// implemented either. but its a starting point to show whats inside
// a rudimentary buzz sample playback machine.
//
//

#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include "../MachineInterface.h"
#include "../dsplib/dsplib.h"

#pragma optimize ("awy", on)

// This basically tells us a sample isn't 16-bit.. and can be 24-bit or 32-bit
#define WF_NOT16		4

#define WFE_32BITF		1
#define WFE_32BITI		2
#define WFE_24BIT		3

#define MAX_TRACKS					8
#define undenormalise(sample) if(((*(unsigned int*)&sample)&0x7f800000)==0) sample=0.0f

//___________________________________________________________________
// 
// Parameters and gruelling machine paperwork starts here
//___________________________________________________________________

CMachineParameter const paraNote = {
	pt_note,
	"Note",
	"Note",
	NOTE_MIN,
	NOTE_OFF,
	NOTE_NO,
	MPF_TICK_ON_EDIT,
	0
};

CMachineParameter const paraWaveNumber = {
	pt_byte,
	"Wave Number",
	"Wave Number",
	WAVE_MIN,
	WAVE_MAX,
	WAVE_NO,
	MPF_WAVE,
	0
};

CMachineParameter const paraVolume = {
	pt_byte,
	"Volume",
	"Volume (0 = 0%, 80 = 100%, FE = ~198%)",
	0,
	0xFE,
	0xFF,
	0,
	0
};

CMachineParameter const paraEffect1 = {
	pt_byte,
	"Effect 1",
	"Effect 1",
	0,
	0xFE,
	0xFF,
	0,
	0
};

CMachineParameter const paraEffectData1 = {
	pt_byte,
	"Effect Data 1",
	"Effect Data 1",
	0,
	0xFE,
	0xFF,
	0,
	0
};

CMachineParameter const paraEffect2 = {
	pt_byte,
	"Effect 2",
	"Effect 2",
	0,
	0xFE,
	0xFF,
	0,
	0
};

CMachineParameter const paraEffectData2 = {
	pt_byte,
	"Effect Data 2",
	"Effect Data 2",
	0,
	0xFE,
	0xFF,
	0,
	0
};

CMachineParameter const *pParameters[] = {
	&paraNote,
	&paraWaveNumber,
	&paraVolume,
	&paraEffect1,
	&paraEffectData1,
	&paraEffect2,
	&paraEffectData2

};

CMachineAttribute const attrMIDIChannel = 
{
	"MIDI Channel",
	0,
	16,
	0
};

CMachineAttribute const attrMIDIVelocity = 
{
	"MIDI Velocity",
	0,
	15,
	2
};

CMachineAttribute const attrMIDITranspose = 
{
	"MIDI Transpose (-24..24 smt)",
	0,
	48,
	24
};

CMachineAttribute const attrMIDIRecordable = 
{
	"MIDI Recording Enabled",
	0,
	1,
	1
};

CMachineAttribute const *pAttributes[] =
{
	&attrMIDIChannel,
	&attrMIDIVelocity,
	&attrMIDITranspose,
	&attrMIDIRecordable
};

#pragma pack(1)

class tvals
{
public:
	byte note;
	byte number;
	byte volume;
	byte effect1;
	byte effectdata1;
	byte effect2;
	byte effectdata2;
};

class avals
{
public:
	int midichannel;
	int midivelocity;
	int miditranspose;
	int midirecord;
};

#pragma pack()

CMachineInfo const MacInfo =
{

	MT_GENERATOR,
	MI_VERSION,
	MIF_PLAYS_WAVES | MIF_MONO_TO_STEREO,
	1,
	MAX_TRACKS,
	0,
	7,
	pParameters,
	4,
	pAttributes,
	"CyanPhase MiniTrk",
	"MiniTrk",
	"Edward L. Blake",
	NULL
};

class mi;

#pragma pack(8)

class CTrack
{
public:
	void Tick(tvals const &tv);
	void Stop();
	inline void Generate(float *psamples, int numsamples);
public:
	double cursamp;
	double veloread;
	double loopstart;
	double loopend;
	float xm1L, x0L, x1L, x2L;
	float xm1R, x0R, x1R, x2R;
	float wvolume;
	float volume;
	int number;
	int playing;
	int gotsomething;
	int mynote;
	int mybase;
	int isstereo, isloop, isbidirloop;
	int sampleenc;
	byte CurrentNote;

	short *sampleset0; // samples accessible as arrays in short (int16) form
	float *sampleset1; // samples accessible as arrays in float form
	long  *sampleset2; // samples accessible as arrays in long (int32) form
	// sampleset3 is 24 bit...


	mi *pmi;
};

#pragma pack()

class mi : public CMachineInterface
{
public:
	mi();
	virtual ~mi();

	virtual void Tick();
	virtual void Stop();

	virtual void Init(CMachineDataInput * const pi);
	virtual bool WorkMonoToStereo(float *pin, float *pout, int numsamples, int const mode);
	virtual char const *mi::DescribeValue(int const param, int const value);
	virtual void SetNumTracks(int const n) { numTracks = n; }
	virtual int AllocateTrack(CSequence *pseq, int note);
	virtual void MidiNoteOff(int c, CSequence *pseq, int notedelay);
	virtual void MidiNote(int const channel, int const value, int const velocity);

public:
	CMachine *ThisMachine;
	int numTracks;
	CTrack Tracks[MAX_TRACKS];
	tvals tval[MAX_TRACKS];
	avals aval;

};

DLL_EXPORTS

mi::mi()
{
	TrackVals = tval;
	AttrVals = (int *)&aval;
}

mi::~mi() 
{

}

//___________________________________________________________________
// 
// Track Level functions start here
//___________________________________________________________________

void CTrack::Tick(tvals const &tv)
{
	if (tv.effect1 != 0xFF) {

	};

	if (tv.effectdata1 != 0xFF) {

	};

	if (tv.effect2 != 0xFF) {

	};

	if (tv.effectdata2 != 0xFF) {

	};

	//if (tv.volume != 0xFF) volume = (double)(tv.volume / 128.0f);
	if (tv.volume != 0xFF) volume = (float)(tv.volume / 128.0f);
	if (tv.number != WAVE_NO) {
		playing = 0;
		cursamp = 0.0;
		number = tv.number;
	};

	if (tv.note != NOTE_NO)
	{
		if (pmi->pCB->GetWaveLevel(number,0) == NULL) {
			playing = 0;
			cursamp = 0.0;
		} else {
			if (tv.note != NOTE_OFF) {
				playing = 1;
				cursamp = 0.0;

				CurrentNote = tv.note;
				mynote = tv.note;
				int l_Note = ((mynote >> 4) * 12) + ((mynote & 15) - 1);
				mybase = pmi->pCB->GetNearestWaveLevel(number,mynote)->RootNote;
				int l_Moo = ((mybase >> 4) * 12) + ((mybase & 15) - 1);
				int N = l_Note - l_Moo;

				if ((pmi->pCB->GetWave(number)->Flags & WF_LOOP) > 0) isloop = 1; else isloop = 0;
				if ((pmi->pCB->GetWave(number)->Flags & WF_BIDIR_LOOP) > 0) isbidirloop = 1; else isbidirloop = 0;
				if ((pmi->pCB->GetWave(number)->Flags & WF_NOT16) > 0) {
					switch (pmi->pCB->GetNearestWaveLevel(number,mynote)->pSamples[0]) {
					case WFE_32BITF: {
						sampleenc = WFE_32BITF;
						sampleset1 = (float *)(&pmi->pCB->GetNearestWaveLevel(number,mynote)->pSamples[4]);
						loopstart = (double)((pmi->pCB->GetNearestWaveLevel(number,mynote)->LoopStart - 4) / 2);
						loopend = (double)((pmi->pCB->GetNearestWaveLevel(number,mynote)->LoopEnd - 4) / 2);
						break; }
					case WFE_32BITI: {
						sampleenc = WFE_32BITI;
						sampleset2 = (long *)(&pmi->pCB->GetNearestWaveLevel(number,mynote)->pSamples[4]);
						loopstart = (double)((pmi->pCB->GetNearestWaveLevel(number,mynote)->LoopStart - 4) / 2);
						loopend = (double)((pmi->pCB->GetNearestWaveLevel(number,mynote)->LoopEnd - 4) / 2);
						break; }
					case WFE_24BIT:
						// not implemented yet.. I'm lazy, you can code
						// 24bit support based on the general format above
						// data starts after 4th short.. however it is not
						// 16bit aligned so you'll have to devise an efficient
						// way to play [ 3 Bytes L ] [ 3 Bytes R ] [ 3 Bytes L ] ...
						sampleenc = WFE_24BIT;
						break;
					}
				} else {
					sampleenc = 0;
					sampleset0 = (short *)(&pmi->pCB->GetNearestWaveLevel(number,mynote)->pSamples[0]);
					loopstart = (double)(pmi->pCB->GetNearestWaveLevel(number,mynote)->LoopStart);
					loopend = (double)(pmi->pCB->GetNearestWaveLevel(number,mynote)->LoopEnd);
				}

				wvolume = pmi->pCB->GetWave(number)->Volume;

				veloread = (((double)pmi->pCB->GetNearestWaveLevel(number,mynote)->SamplesPerSec) / (double)pmi->pMasterInfo->SamplesPerSec) * pow(2.0,(N / 12.0));

				if ((pmi->pCB->GetWave(number)->Flags & WF_STEREO) > 0) {
					isstereo = 1;
					switch (sampleenc) {
					case WFE_32BITF:
						xm1L = 0.0f;
						x0L = sampleset1[0] * 32767.0f;
						x1L = sampleset1[2] * 32767.0f;
						x2L = sampleset1[4] * 32767.0f;
						xm1R = 0.0f;
						x0R = sampleset1[1] * 32767.0f;
						x1R = sampleset1[3] * 32767.0f;
						x2R = sampleset1[5] * 32767.0f;
						break;
					case WFE_32BITI:
						xm1L = 0.0f;
						x0L = sampleset2[0] / 65538.0f;
						x1L = sampleset2[2] / 65538.0f;
						x2L = sampleset2[4] / 65538.0f;
						xm1R = 0.0f;
						x0R = sampleset2[1] / 65538.0f;
						x1R = sampleset2[3] / 65538.0f;
						x2R = sampleset2[5] / 65538.0f;
						break;
					case WFE_24BIT:
						// not implemented yet.. I'm lazy, you can code
						// 24bit support based on the general format above
						// data starts after 4th short.. however it is not
						// 16bit aligned so you'll have to devise an efficient
						// way to play [ 3 Bytes L ] [ 3 Bytes R ] [ 3 Bytes L ] ...
						break;
					case 0:
						xm1L = 0.0f;
						x0L = sampleset0[0];
						x1L = sampleset0[2];
						x2L = sampleset0[4];
						xm1R = 0.0f;
						x0R = sampleset0[1];
						x1R = sampleset0[3];
						x2R = sampleset0[5];
					}
				} else {
					isstereo = 0;
					switch (sampleenc) {
					case WFE_32BITF:
						xm1L = 0.0f;
						x0L = sampleset1[0] * 32767.0f;
						x1L = sampleset1[1] * 32767.0f;
						x2L = sampleset1[2] * 32767.0f;
						break;
					case WFE_32BITI:
						xm1L = 0.0f;
						x0L = sampleset2[0] / 65538.0f;
						x1L = sampleset2[1] / 65538.0f;
						x2L = sampleset2[2] / 65538.0f;
						break;
					case WFE_24BIT:
						// not implemented yet.. I'm lazy, you can code
						// 24bit support based on the general format above
						// data starts after 4th short.. however it is not
						// 16bit aligned so you'll have to devise an efficient
						// way to play [ 3 Bytes L ] [ 3 Bytes R ] [ 3 Bytes L ] ...
						break;
					case 0:
						xm1L = 0.0f;
						x0L = sampleset0[0];
						x1L = sampleset0[1];
						x2L = sampleset0[2];
					}
				};
			} else {
				playing = 0;
				cursamp = 0.0;
			};
		};
	};
}

inline void CTrack::Generate(float *psamples, int numsamples)
{
	double const ad2i = (1.5 * (1 << 26) * (1 << 26));
	double magicres;
	double curfrac;
	int rVal;
	float in;
	unsigned int ns;
	unsigned int ucursamp;
	int i;

	if (playing == 0) {
		gotsomething = 0;
		return;
	};

	gotsomething = 1;
	
	if (isloop == 1) {

		wvolume = (pmi->pCB->GetWave(number)->Volume);
		CWaveLevel const *pLevel = pmi->pCB->GetNearestWaveLevel(number, mynote);

		if (isstereo == 1) {
			ns = (unsigned int)(loopend * 2.0 - 7.0);

			for (i = 0; i < numsamples*2; i++) {
				if (playing == 1) {
					cursamp += veloread;
					if (veloread < 0.0) {
						if (cursamp < loopstart) {
							veloread = -veloread;
						};
					};
					magicres = (cursamp - 0.5) + ad2i;
					rVal =  *(int *)&magicres;
					ucursamp = (unsigned int)(rVal) * 2u;
					curfrac = cursamp - (double)rVal;

					if (ucursamp < ns) {

						switch (sampleenc) {
						case WFE_32BITF:
							xm1L = sampleset1[ucursamp - 2u] * 32767.0f;
							x0L = sampleset1[ucursamp] * 32767.0f;
							x1L = sampleset1[ucursamp + 2u] * 32767.0f;
							x2L = sampleset1[ucursamp + 4u] * 32767.0f;
							break;
						case WFE_32BITI:
							xm1L = sampleset2[ucursamp - 2u] / 65538.0f;
							x0L = sampleset2[ucursamp] / 65538.0f;
							x1L = sampleset2[ucursamp + 2u] / 65538.0f;
							x2L = sampleset2[ucursamp + 4u] / 65538.0f;
							break;
						case WFE_24BIT:
							// not implemented yet.. I'm lazy, you can code
							// 24bit support based on the general format above
							// data starts after 4th short.. however it is not
							// 16bit aligned so you'll have to devise an efficient
							// way to play [ 3 Bytes L ] [ 3 Bytes R ] [ 3 Bytes L ] ...
							break;
						case 0:
							xm1L = (float)(sampleset0[ucursamp - 2u]);
							x0L = (float)(sampleset0[ucursamp]);
							x1L = (float)(sampleset0[ucursamp + 2u]);
							x2L = (float)(sampleset0[ucursamp + 4u]);
						}

						in = ((((3.0f * (x0L - x1L) - xm1L + x2L) * 0.5f * curfrac) + 2.0f * x1L + xm1L - (5.0f * x0L + x2L) * 0.5f) * curfrac + (x1L - xm1L) * 0.5f) * curfrac + x0L;

						*psamples++ = (in * wvolume * volume);

						switch (sampleenc) {
						case WFE_32BITF:
							xm1R = sampleset1[ucursamp - 1u] * 32767.0f;
							x0R = sampleset1[ucursamp + 1u] * 32767.0f;
							x1R = sampleset1[ucursamp + 3u] * 32767.0f;
							x2R = sampleset1[ucursamp + 5u] * 32767.0f;
							break;
						case WFE_32BITI:
							xm1R = sampleset2[ucursamp - 1u] / 65538.0f;
							x0R = sampleset2[ucursamp + 1u] / 65538.0f;
							x1R = sampleset2[ucursamp + 3u] / 65538.0f;
							x2R = sampleset2[ucursamp + 5u] / 65538.0f;
							break;
						case WFE_24BIT:
							// not implemented yet.. I'm lazy, you can code
							// 24bit support based on the general format above
							// data starts after 4th short.. however it is not
							// 16bit aligned so you'll have to devise an efficient
							// way to play [ 3 Bytes L ] [ 3 Bytes R ] [ 3 Bytes L ] ...
							break;
						case 0:
							xm1R = (float)(sampleset0[ucursamp - 1u]);
							x0R = (float)(sampleset0[ucursamp + 1u]);
							x1R = (float)(sampleset0[ucursamp + 3u]);
							x2R = (float)(sampleset0[ucursamp + 5u]);
						}

						i++;

						in = ((((3.0f * (x0R - x1R) - xm1R + x2R) * 0.5f * curfrac) + 2.0f * x1R + xm1R - (5.0f * x0R + x2R) * 0.5f) * curfrac + (x1R - xm1R) * 0.5f) * curfrac + x0R;

						*psamples++ = (in * wvolume * volume);
					}
					else
					{
						if (isbidirloop == 1) {
							if (veloread > 0.0) {
								veloread = -veloread;
							};
						} else {
							cursamp = loopstart;
						};

						switch (sampleenc) {
						case WFE_32BITF:
							xm1L = sampleset1[ucursamp - 2u] * 32767.0f;
							x0L = sampleset1[ucursamp] * 32767.0f;
							x1L = sampleset1[ucursamp + 2u] * 32767.0f;
							x2L = sampleset1[ucursamp + 4u] * 32767.0f;
							break;
						case WFE_32BITI:
							xm1L = sampleset2[ucursamp - 2u] / 65538.0f;
							x0L = sampleset2[ucursamp] / 65538.0f;
							x1L = sampleset2[ucursamp + 2u] / 65538.0f;
							x2L = sampleset2[ucursamp + 4u] / 65538.0f;
							break;
						case WFE_24BIT:
							// not implemented yet.. I'm lazy, you can code
							// 24bit support based on the general format above
							// data starts after 4th short.. however it is not
							// 16bit aligned so you'll have to devise an efficient
							// way to play [ 3 Bytes L ] [ 3 Bytes R ] [ 3 Bytes L ] ...
							break;
						case 0:
							xm1L = (float)(sampleset0[ucursamp - 2u]);
							x0L = (float)(sampleset0[ucursamp]);
							x1L = (float)(sampleset0[ucursamp + 2u]);
							x2L = (float)(sampleset0[ucursamp + 4u]);
						}

						in = ((((3.0f * (x0L - x1L) - xm1L + x2L) * 0.5f * curfrac) + 2.0f * x1L + xm1L - (5.0f * x0L + x2L) * 0.5f) * curfrac + (x1L - xm1L) * 0.5f) * curfrac + x0L;

						*psamples++ = (in * wvolume * volume);

						switch (sampleenc) {
						case WFE_32BITF:
							xm1R = sampleset1[ucursamp - 1u] * 32767.0f;
							x0R = sampleset1[ucursamp + 1u] * 32767.0f;
							x1R = sampleset1[ucursamp + 3u] * 32767.0f;
							x2R = sampleset1[ucursamp + 5u] * 32767.0f;
							break;
						case WFE_32BITI:
							xm1R = sampleset2[ucursamp - 1u] / 65538.0f;
							x0R = sampleset2[ucursamp + 1u] / 65538.0f;
							x1R = sampleset2[ucursamp + 3u] / 65538.0f;
							x2R = sampleset2[ucursamp + 5u] / 65538.0f;
							break;
						case WFE_24BIT:
							// not implemented yet.. I'm lazy, you can code
							// 24bit support based on the general format above
							// data starts after 4th short.. however it is not
							// 16bit aligned so you'll have to devise an efficient
							// way to play [ 3 Bytes L ] [ 3 Bytes R ] [ 3 Bytes L ] ...
							break;
						case 0:
							xm1R = (float)(sampleset0[ucursamp - 1u]);
							x0R = (float)(sampleset0[ucursamp + 1u]);
							x1R = (float)(sampleset0[ucursamp + 3u]);
							x2R = (float)(sampleset0[ucursamp + 5u]);
						}

						i++;

						in = ((((3.0f * (x0R - x1R) - xm1R + x2R) * 0.5f * curfrac) + 2.0f * x1R + xm1R - (5.0f * x0R + x2R) * 0.5f) * curfrac + (x1R - xm1R) * 0.5f) * curfrac + x0R;

						*psamples++ = (in * wvolume * volume);
					};			
				}
				else
				{
					*psamples++ = 0.0f;
					i++;
					*psamples++ = 0.0f;
				};
			};
		} else {
			ns = (unsigned int)(loopend - 3);


			for (i = 0; i < numsamples; i++) {
				in = 0.0f;
				if (playing == 1) {
					cursamp += veloread;
					if (veloread < 0.0) {
						if (cursamp < loopstart) {
							veloread = -veloread;
						};
					};
					magicres = (cursamp - 0.5) + ad2i;
					rVal =  *(int *)&magicres;
					ucursamp = (unsigned int)(rVal);
					curfrac = cursamp - (double)rVal;

					if (ucursamp < ns) {
						switch (sampleenc) {
						case WFE_32BITF:
							xm1L = sampleset1[ucursamp - 1u] * 32767.0f;
							x0L = sampleset1[ucursamp] * 32767.0f;
							x1L = sampleset1[ucursamp + 1u] * 32767.0f;
							x2L = sampleset1[ucursamp + 2u] * 32767.0f;
							break;
						case WFE_32BITI:
							xm1L = sampleset2[ucursamp - 1u] / 65538.0f;
							x0L = sampleset2[ucursamp] / 65538.0f;
							x1L = sampleset2[ucursamp + 1u] / 65538.0f;
							x2L = sampleset2[ucursamp + 2u] / 65538.0f;
							break;
						case WFE_24BIT:
							// not implemented yet.. I'm lazy, you can code
							// 24bit support based on the general format above
							// data starts after 4th short.. however it is not
							// 16bit aligned so you'll have to devise an efficient
							// way to play [ 3 Bytes L ] [ 3 Bytes R ] [ 3 Bytes L ] ...
							break;
						case 0:
							xm1L = (float)(sampleset0[ucursamp - 1u]);
							x0L = (float)(sampleset0[ucursamp]);
							x1L = (float)(sampleset0[ucursamp + 1u]);
							x2L = (float)(sampleset0[ucursamp + 2u]);
						}

						in = ((((3.0f * (x0L - x1L) - xm1L + x2L) * 0.5f * curfrac) + 2.0f * x1L + xm1L - (5.0f * x0L + x2L) * 0.5f) * curfrac + (x1L - xm1L) * 0.5f) * curfrac + x0L;
					}
					else
					{
						if (isbidirloop == 1) {
							if (veloread > 0.0) {
								veloread = -veloread;
							};
						} else {
							cursamp = loopstart;
						};
						switch (sampleenc) {
						case WFE_32BITF:
							xm1L = sampleset1[ucursamp - 1u] * 32767.0f;
							x0L = sampleset1[ucursamp] * 32767.0f;
							x1L = sampleset1[ucursamp + 1u] * 32767.0f;
							x2L = sampleset1[ucursamp + 2u] * 32767.0f;
							break;
						case WFE_32BITI:
							xm1L = sampleset2[ucursamp - 1u] / 65538.0f;
							x0L = sampleset2[ucursamp] / 65538.0f;
							x1L = sampleset2[ucursamp + 2u] / 65538.0f;
							x2L = sampleset2[ucursamp + 2u] / 65538.0f;
							break;
						case WFE_24BIT:
							// not implemented yet.. I'm lazy, you can code
							// 24bit support based on the general format above
							// data starts after 4th short.. however it is not
							// 16bit aligned so you'll have to devise an efficient
							// way to play [ 3 Bytes L ] [ 3 Bytes R ] [ 3 Bytes L ] ...
							break;
						case 0:
							xm1L = (float)(sampleset0[ucursamp - 1u]);
							x0L = (float)(sampleset0[ucursamp]);
							x1L = (float)(sampleset0[ucursamp + 1u]);
							x2L = (float)(sampleset0[ucursamp + 2u]);
						}

						in = ((((3.0f * (x0L - x1L) - xm1L + x2L) * 0.5f * curfrac) + 2.0f * x1L + xm1L - (5.0f * x0L + x2L) * 0.5f) * curfrac + (x1L - xm1L) * 0.5f) * curfrac + x0L;
					};				
				}

				*psamples++ = (in * wvolume * volume);
				*psamples++ = (in * wvolume * volume);
			};
		};

	} else {

		wvolume = (pmi->pCB->GetWave(number)->Volume);
		CWaveLevel const *pLevel = pmi->pCB->GetNearestWaveLevel(number, mynote);

		if (isstereo == 1) {
			switch (sampleenc) {
			case WFE_32BITF:
				ns = (unsigned int)((pLevel->numSamples-4)/2 * 2 - 5);
				break;
			case WFE_32BITI:
				ns = (unsigned int)((pLevel->numSamples-4)/2 * 2 - 5);
				break;
			case WFE_24BIT:
				// not implemented yet.. I'm lazy, you can code
				// 24bit support based on the general format above
				// data starts after 4th short.. however it is not
				// 16bit aligned so you'll have to devise an efficient
				// way to play [ 3 Bytes L ] [ 3 Bytes R ] [ 3 Bytes L ] ...
				//ns = (unsigned int)((pLevel->numSamples-4) * 2 - 5);
				break;
			case 0:
				ns = (unsigned int)(pLevel->numSamples * 2 - 5);
			}

			for (i = 0; i < numsamples*2; i++) {
				if (playing == 1) {
					cursamp += veloread;
					magicres = (cursamp - 0.5) + ad2i;
					rVal =  *(int *)&magicres;
					ucursamp = (unsigned int)(rVal) * 2u;
					curfrac = cursamp - (double)rVal;

					if (ucursamp < ns) {

						switch (sampleenc) {
						case WFE_32BITF:
							xm1L = sampleset1[ucursamp - 2u] * 32767.0f;
							x0L = sampleset1[ucursamp] * 32767.0f;
							x1L = sampleset1[ucursamp + 2u] * 32767.0f;
							x2L = sampleset1[ucursamp + 4u] * 32767.0f;
							break;
						case WFE_32BITI:
							xm1L = sampleset2[ucursamp - 2u] / 65538.0f;
							x0L = sampleset2[ucursamp] / 65538.0f;
							x1L = sampleset2[ucursamp + 2u] / 65538.0f;
							x2L = sampleset2[ucursamp + 4u] / 65538.0f;
							break;
						case WFE_24BIT:
							// not implemented yet.. I'm lazy, you can code
							// 24bit support based on the general format above
							// data starts after 4th short.. however it is not
							// 16bit aligned so you'll have to devise an efficient
							// way to play [ 3 Bytes L ] [ 3 Bytes R ] [ 3 Bytes L ] ...
							break;
						case 0:
							xm1L = (float)(sampleset0[ucursamp - 2u]);
							x0L = (float)(sampleset0[ucursamp]);
							x1L = (float)(sampleset0[ucursamp + 2u]);
							x2L = (float)(sampleset0[ucursamp + 4u]);
						}

						in = ((((3.0f * (x0L - x1L) - xm1L + x2L) * 0.5f * curfrac) + 2.0f * x1L + xm1L - (5.0f * x0L + x2L) * 0.5f) * curfrac + (x1L - xm1L) * 0.5f) * curfrac + x0L;

						*psamples++ = (in * wvolume * volume);

						switch (sampleenc) {
						case WFE_32BITF:
							xm1R = sampleset1[ucursamp - 1u] * 32767.0f;
							x0R = sampleset1[ucursamp + 1u] * 32767.0f;
							x1R = sampleset1[ucursamp + 3u] * 32767.0f;
							x2R = sampleset1[ucursamp + 5u] * 32767.0f;
							break;
						case WFE_32BITI:
							xm1R = sampleset2[ucursamp - 1u] / 65538.0f;
							x0R = sampleset2[ucursamp + 1u] / 65538.0f;
							x1R = sampleset2[ucursamp + 3u] / 65538.0f;
							x2R = sampleset2[ucursamp + 5u] / 65538.0f;
							break;
						case WFE_24BIT:
							// not implemented yet.. I'm lazy, you can code
							// 24bit support based on the general format above
							// data starts after 4th short.. however it is not
							// 16bit aligned so you'll have to devise an efficient
							// way to play [ 3 Bytes L ] [ 3 Bytes R ] [ 3 Bytes L ] ...
							break;
						case 0:
							xm1R = (float)(sampleset0[ucursamp - 1u]);
							x0R = (float)(sampleset0[ucursamp + 1u]);
							x1R = (float)(sampleset0[ucursamp + 3u]);
							x2R = (float)(sampleset0[ucursamp + 5u]);
						}

						i++;

						in = ((((3.0f * (x0R - x1R) - xm1R + x2R) * 0.5f * curfrac) + 2.0f * x1R + xm1R - (5.0f * x0R + x2R) * 0.5f) * curfrac + (x1R - xm1R) * 0.5f) * curfrac + x0R;

						*psamples++ = (in * wvolume * volume);
					}
					else
					{
						playing = 0;
						*psamples++ = 0.0f;
						i++;
						*psamples++ = 0.0f;
					};			
				}
				else
				{
					*psamples++ = 0.0f;
					i++;
					*psamples++ = 0.0f;
				};
			};
		} else {
			switch (sampleenc) {
			case WFE_32BITF:
				ns = (unsigned int)((pLevel->numSamples - 4) /2 - 2);
				break;
			case WFE_32BITI:
				ns = (unsigned int)((pLevel->numSamples - 4)/2 - 2);
				break;
			case WFE_24BIT:
				// not implemented yet.. I'm lazy, you can code
				// 24bit support based on the general format above
				// data starts after 4th short.. however it is not
				// 16bit aligned so you'll have to devise an efficient
				// way to play [ 3 Bytes L ] [ 3 Bytes R ] [ 3 Bytes L ] ...
				//ns = (unsigned int)((pLevel->numSamples - 4) / 2 - 2);
				break;
			case 0:
				ns = (unsigned int)(pLevel->numSamples - 2);
			}


			for (i = 0; i < numsamples; i++) {
				in = 0.0f;
				if (playing == 1) {
					cursamp += veloread;
					magicres = (cursamp - 0.5) + ad2i;
					rVal =  *(int *)&magicres;
					ucursamp = (unsigned int)(rVal);
					curfrac = cursamp - (double)rVal;

					if (ucursamp < ns) {
						switch (sampleenc) {
						case WFE_32BITF:
							xm1L = sampleset1[ucursamp - 1u] * 32767.0f;
							x0L = sampleset1[ucursamp] * 32767.0f;
							x1L = sampleset1[ucursamp + 1u] * 32767.0f;
							x2L = sampleset1[ucursamp + 2u] * 32767.0f;
							break;
						case WFE_32BITI:
							xm1L = sampleset2[ucursamp - 1u] / 65538.0f;
							x0L = sampleset2[ucursamp] / 65538.0f;
							x1L = sampleset2[ucursamp + 1u] / 65538.0f;
							x2L = sampleset2[ucursamp + 2u] / 65538.0f;
							break;
						case WFE_24BIT:
							// not implemented yet.. I'm lazy, you can code
							// 24bit support based on the general format above
							// data starts after 4th short.. however it is not
							// 16bit aligned so you'll have to devise an efficient
							// way to play [ 3 Bytes L ] [ 3 Bytes R ] [ 3 Bytes L ] ...
							break;
						case 0:
							xm1L = (float)(sampleset0[ucursamp - 1u]);
							x0L = (float)(sampleset0[ucursamp]);
							x1L = (float)(sampleset0[ucursamp + 1u]);
							x2L = (float)(sampleset0[ucursamp + 2u]);
						}

						in = ((((3.0f * (x0L - x1L) - xm1L + x2L) * 0.5f * curfrac) + 2.0f * x1L + xm1L - (5.0f * x0L + x2L) * 0.5f) * curfrac + (x1L - xm1L) * 0.5f) * curfrac + x0L;
					}
					else
					{
						playing = 0;
					//	in = 0.0f;

					};				
				}

				*psamples++ = (in * wvolume * volume);
				*psamples++ = (in * wvolume * volume);
			};
		};
	};
}

void CTrack::Stop() {
	playing = 0;
}


//___________________________________________________________________
// 
// Machine Level functions start here
//___________________________________________________________________

void mi::Init(CMachineDataInput * const pi)
{
	ThisMachine = pCB->GetThisMachine();

	for (int i = 0; i < MAX_TRACKS; i++) {
		
		Tracks[i].playing = 0;
		Tracks[i].gotsomething = 0;
		Tracks[i].cursamp = 0.0;
		Tracks[i].number = 1;
		Tracks[i].volume = 1.0;

		Tracks[i].isstereo = 0;
		Tracks[i].isloop = 0;
		Tracks[i].isbidirloop = 0;
		
		Tracks[i].pmi = this;
	};
}

void mi::Tick()
{
	for (int i = 0; i < numTracks; i++)
		Tracks[i].Tick(tval[i]);
}

bool mi::WorkMonoToStereo(float *pin, float *pout, int numsamples, int const mode)
{
	bool gotsomething = false;
	bool firstone = true;
	int i;

	float *paux = pCB->GetAuxBuffer();

	for (i = 0; i < numTracks; i++){
		Tracks[i].Generate(paux, numsamples);
		if (Tracks[i].gotsomething == 1) {
			gotsomething = true;
			if (firstone == true) {
				DSP_Copy(pout, paux, numsamples*2);
				firstone = false;
			} else {
				DSP_Add(pout, paux, numsamples*2);
			};
		}
	}

	return gotsomething;
}

void mi::Stop()
{
	for (int c = 0; c < numTracks; c++)
		Tracks[c].Stop();
}

char const *mi::DescribeValue(int const param, int const value)
{
	static char txt[16];

	switch(param)
	{
	case 0:				// Note
		return NULL;
		break;
	case 1:				// Wave Number
		return NULL;
		break;
	case 2:				// Volume
		sprintf(txt,"%.2f s",value / 100.0f);
		return txt;
		break;
	case 3:				// Effect
	case 5:
		switch (value)
		{
		case 0: return "Arpeggio"; break; 
		case 1: return "Slide up"; break;
		case 2: return "Slide down"; break;
		case 3: return "Tone portamento"; break;
		case 4: return "Vibrato"; break;
		case 5: return "Slide panning"; break;
		case 6: return "Autopan"; break;
		case 7: return "Tremolo"; break;
		case 8: return "Set panning position"; break;
		case 9: return "Offset"; break;
		case 10: return "Volume slide"; break;
		case 15: return "Subdivide amount"; break;
		case 16: return "Probability"; break;
		case 17: return "Loop fit"; break;
		case 18: return "Loop fit with tracking"; break;
		case 19: return "Auto shuffle"; break;
		case 20: return "Randomize volume"; break;
		case 21: return "Random delay"; break;
		case 22: return "Randomize pitch"; break;
		case 23: return "Harmonic play"; break;
		case 24: return "Combined note delay and cut"; break;
		case 25: return "Sustain pedal"; break;
		case 220: return "Note release"; break;
		case 225: return "Fine slide up"; break;
		case 226: return "Fine slide down"; break;
		case 228: return "Set vibrato type"; break;
		case 229: return "Set finetune"; break;
		case 230: return "Set panning type"; break;
		case 231: return "Set tremolo type"; break;
		case 232: return "Reverse direction"; break;
		case 233: return "Retrig sample"; break;
		case 234: return "Fine volume slide up"; break;
		case 235: return "Fine volume slide down"; break;
		case 236: return "Note cutoff"; break;
		case 237: return "Note delay"; break;
		case 238: return "Fine panning slide left"; break;
		case 239: return "Fine panning slide right"; break;
		default: return "Command not implemented"; break;
		}
		break;
	case 4:				// Effect Data
	case 6:
		return NULL;
		break;
	default: return NULL;
		break;
	};
}



int mi::AllocateTrack(CSequence *pseq, int note)
{
	
	for (int c = 0; c < numTracks; c++)
	{
		if (Tracks[c].CurrentNote == note)
		{
			if (pseq != NULL)
			{
				byte *pdata = (byte *)pCB->GetPlayingRow(pseq, 2, c);
				if (*pdata == NOTE_NO)
					return c;
			}
		}
	}

	int best = -1;

	for (c = 0; c < numTracks; c++)
	{
		if (Tracks[c].playing == 0)
		{
			if (pseq != NULL)
			{
				byte *pdata = (byte *)pCB->GetPlayingRow(pseq, 2, c);
				if (*pdata == NOTE_NO)
					return c;
				
			}

			if (best == -1)
				best = c;
		}
	}

	if (best != -1)
		return best;

	return -1;
}

void mi::MidiNoteOff(int c, CSequence *pseq, int notedelay)
{
	Tracks[c].playing = 0;
	Tracks[c].cursamp = 0.0;

	if (pseq != NULL)
	{
		byte *pdata = (byte *)pCB->GetPlayingRow(pseq, 2, c);

		if (pdata[3] == NOTE_NO) {
			pdata[3] = NOTE_OFF;
		};
	}
}


void mi::MidiNote(int const channel, int const value, int const velocity)
{
	if (aval.midichannel == channel)
	{
		if (value / 12 > 9) return;	
		bool dontrecord = false;
		
		int myvelocity;
		CSequence *pseq;

		int reval = value + (aval.miditranspose - 24);
		byte n = (((reval / 12)-1) << 4) | ((reval % 12) + 1);

		if (aval.midirecord == 1) {
			int stateflags = pCB->GetStateFlags();
			
			if (stateflags & SF_PLAYING && stateflags & SF_RECORDING)
				pseq = pCB->GetPlayingSequence(ThisMachine);
			else 
				pseq = NULL;
		} else {
			pseq = NULL;
		};

		int notedelay = pMasterInfo->PosInTick * 24 / pMasterInfo->SamplesPerTick;

		if (velocity > 0)
		{
			int c = AllocateTrack(pseq, n);

			if (c == -1) {
				c = 0;
				dontrecord = true;
			};

			Tracks[c].playing = 1;
			Tracks[c].cursamp = 0.0;

			if ((pCB->GetWave(Tracks[c].number)->Flags & WF_LOOP) > 0) Tracks[c].isloop = 1; else Tracks[c].isloop = 0;
			if ((pCB->GetWave(Tracks[c].number)->Flags & WF_BIDIR_LOOP) > 0) Tracks[c].isbidirloop = 1; else Tracks[c].isbidirloop = 0;

			Tracks[c].mynote = n;
			int l_Note = ((Tracks[c].mynote >> 4) * 12) + ((Tracks[c].mynote & 15) - 1);
			Tracks[c].mybase = pCB->GetNearestWaveLevel(Tracks[c].number,Tracks[c].mynote)->RootNote;
			int l_Moo = ((Tracks[c].mybase >> 4) * 12) + ((Tracks[c].mybase & 15) - 1);
			int N = l_Note - l_Moo;

			if ((pCB->GetWave(Tracks[c].number)->Flags & WF_NOT16) > 0) {
				switch (pCB->GetNearestWaveLevel(Tracks[c].number,Tracks[c].mynote)->pSamples[0]) {
				case WFE_32BITF: {
					Tracks[c].sampleenc = WFE_32BITF;
					Tracks[c].sampleset1 = (float *)(&pCB->GetNearestWaveLevel(Tracks[c].number,Tracks[c].mynote)->pSamples[4]);
					Tracks[c].loopstart = (double)((pCB->GetNearestWaveLevel(Tracks[c].number,Tracks[c].mynote)->LoopStart - 4) / 2);
					Tracks[c].loopend = (double)((pCB->GetNearestWaveLevel(Tracks[c].number,Tracks[c].mynote)->LoopEnd - 4) / 2);
					break; }
				case WFE_32BITI: {
					Tracks[c].sampleenc = WFE_32BITI;
					Tracks[c].sampleset2 = (long *)(&pCB->GetNearestWaveLevel(Tracks[c].number,Tracks[c].mynote)->pSamples[4]);
					Tracks[c].loopstart = (double)((pCB->GetNearestWaveLevel(Tracks[c].number,Tracks[c].mynote)->LoopStart - 4) / 2);
					Tracks[c].loopend = (double)((pCB->GetNearestWaveLevel(Tracks[c].number,Tracks[c].mynote)->LoopEnd - 4) / 2);
					break; }
				case WFE_24BIT:
					// not implemented yet.. I'm lazy, you can code
					// 24bit support based on the general format above
					// data starts after 4th short.. however it is not
					// 16bit aligned so you'll have to devise an efficient
					// way to play [ 3 Bytes L ] [ 3 Bytes R ] [ 3 Bytes L ] ...
					Tracks[c].sampleenc = WFE_24BIT;
					break;
				}
			} else {
				Tracks[c].sampleenc = 0;
				Tracks[c].sampleset0 = (short *)(&pCB->GetNearestWaveLevel(Tracks[c].number,Tracks[c].mynote)->pSamples[0]);
				Tracks[c].loopstart = (double)(pCB->GetNearestWaveLevel(Tracks[c].number,Tracks[c].mynote)->LoopStart);
				Tracks[c].loopend = (double)(pCB->GetNearestWaveLevel(Tracks[c].number,Tracks[c].mynote)->LoopEnd);
			}


			Tracks[c].wvolume = (pCB->GetWave(Tracks[c].number)->Volume);

			Tracks[c].veloread = ((pCB->GetNearestWaveLevel(Tracks[c].number,Tracks[c].mynote)->SamplesPerSec) / pMasterInfo->SamplesPerSec) * pow(2.0,(N / 12.0));

			if ((pCB->GetWave(Tracks[c].number)->Flags & WF_STEREO) > 0) {
				Tracks[c].isstereo = 1;
				switch (Tracks[c].sampleenc) {
				case WFE_32BITF:
					Tracks[c].xm1L = 0.0f;
					Tracks[c].x0L = Tracks[c].sampleset1[0] * 32767.0f;
					Tracks[c].x1L = Tracks[c].sampleset1[2] * 32767.0f;
					Tracks[c].x2L = Tracks[c].sampleset1[4] * 32767.0f;
					Tracks[c].xm1R = 0.0f;
					Tracks[c].x0R = Tracks[c].sampleset1[1] * 32767.0f;
					Tracks[c].x1R = Tracks[c].sampleset1[3] * 32767.0f;
					Tracks[c].x2R = Tracks[c].sampleset1[5] * 32767.0f;
					break;
				case WFE_32BITI:
					Tracks[c].xm1L = 0.0f;
					Tracks[c].x0L = Tracks[c].sampleset2[0] / 65538.0f;
					Tracks[c].x1L = Tracks[c].sampleset2[2] / 65538.0f;
					Tracks[c].x2L = Tracks[c].sampleset2[4] / 65538.0f;
					Tracks[c].xm1R = 0.0f;
					Tracks[c].x0R = Tracks[c].sampleset2[1] / 65538.0f;
					Tracks[c].x1R = Tracks[c].sampleset2[3] / 65538.0f;
					Tracks[c].x2R = Tracks[c].sampleset2[5] / 65538.0f;
					break;
				case WFE_24BIT:
					// not implemented yet.. I'm lazy, you can code
					// 24bit support based on the general format above
					// data starts after 4th short.. however it is not
					// 16bit aligned so you'll have to devise an efficient
					// way to play [ 3 Bytes L ] [ 3 Bytes R ] [ 3 Bytes L ] ...
					break;
				case 0:
					Tracks[c].xm1L = 0.0;
					Tracks[c].x0L = Tracks[c].sampleset0[0];
					Tracks[c].x1L = Tracks[c].sampleset0[2];
					Tracks[c].x2L = Tracks[c].sampleset0[4];
					Tracks[c].xm1R = 0.0;
					Tracks[c].x0R = Tracks[c].sampleset0[1];
					Tracks[c].x1R = Tracks[c].sampleset0[3];
					Tracks[c].x2R = Tracks[c].sampleset0[5];
				}
			} else {
				Tracks[c].isstereo = 0;
				switch (Tracks[c].sampleenc) {
				case WFE_32BITF:
					Tracks[c].xm1L = 0.0f;
					Tracks[c].x0L = Tracks[c].sampleset1[0] * 32767.0f;
					Tracks[c].x1L = Tracks[c].sampleset1[1] * 32767.0f;
					Tracks[c].x2L = Tracks[c].sampleset1[2] * 32767.0f;
					break;
				case WFE_32BITI:
					Tracks[c].xm1L = 0.0f;
					Tracks[c].x0L = Tracks[c].sampleset2[0] / 65538.0f;
					Tracks[c].x1L = Tracks[c].sampleset2[1] / 65538.0f;
					Tracks[c].x2L = Tracks[c].sampleset2[2] / 65538.0f;
					break;
				case WFE_24BIT:
					// not implemented yet.. I'm lazy, you can code
					// 24bit support based on the general format above
					// data starts after 4th short.. however it is not
					// 16bit aligned so you'll have to devise an efficient
					// way to play [ 3 Bytes L ] [ 3 Bytes R ] [ 3 Bytes L ] ...
					break;
				case 0:
					Tracks[c].xm1L = 0.0;
					Tracks[c].x0L = Tracks[c].sampleset0[0];
					Tracks[c].x1L = Tracks[c].sampleset0[1];
					Tracks[c].x2L = Tracks[c].sampleset0[2];
				}
			};

			Tracks[c].CurrentNote = n;
			switch ((int)(aval.midivelocity))
			{
			case 0: break;
			case 1: // Volume - Light
				myvelocity = velocity * 2;
				if (velocity > 100) myvelocity = 254;
				if (velocity < 0) myvelocity = 0;
				Tracks[c].volume = (float)(myvelocity / 128.0f);
				break; 
			case 2: // Volume - Medium
				myvelocity = velocity + 128;
				if (velocity > 100) myvelocity = 254;
				if (velocity < 0) myvelocity = 0;
				Tracks[c].volume = (float)(myvelocity / 128.0f);
				break; 
			case 3: // Volume - Hard
				myvelocity = (velocity / 2) + 200;
				if (velocity > 100) myvelocity = 254;
				if (velocity < 0) myvelocity = 0;
				Tracks[c].volume = (float)(myvelocity / 128.0f);
				break; 
			case 4: break;
			case 5: break;
			case 6: break;
			case 7: break;
			case 8: break;
			case 9: break;
			case 10: break;
			case 11: break;
			case 12: break;
			case 13: break;
			case 14: break;
			case 15: break;
			default: break;
			}			

			if (aval.midirecord == 1) {
				if (dontrecord == false) {
					if (pseq != NULL)
					{
						byte *pdata = (byte *)pCB->GetPlayingRow(pseq, 2, c);
						switch (((int)aval.midivelocity))
						{
						case 0: pdata[0] = n; break;
						case 1: pdata[0] = n; pdata[2] = myvelocity; break;
						case 2: pdata[0] = n; pdata[2] = myvelocity; break;
						case 3: pdata[0] = n; pdata[2] = myvelocity; break;
						case 4: pdata[0] = n; break;
						case 5: pdata[0] = n; break;
						case 6: pdata[0] = n; break;
						case 7: pdata[0] = n; break;
						case 8: pdata[0] = n; break;
						case 9: pdata[0] = n; break;
						case 10: pdata[0] = n; break;
						case 11: pdata[0] = n; break;
						case 12: pdata[0] = n; break;
						case 13: pdata[0] = n; break;
						case 14: pdata[0] = n; break;
						case 15: pdata[0] = n; break;
						default:
							break;
						}
					}
				};
			};
		}
		else
		{
			for (int c = 0; c < numTracks; c++)
			{
				if (Tracks[c].CurrentNote == n)
				{
					MidiNoteOff(c, pseq, notedelay);
					break;
				}	
			}
		}
	}
}
